#! /usr/bin/env python
# -*- encoding: utf-8 -*-

import codecs
import gzip
import json
import os
import pkg_resources
import random

class Quote(object):
    def __init__(self, author, title, text, text_id):
        self.author = author
        self.text = text
        self.title = title
        self.text_id = text_id


class RandomIterator(object):
    """Random, bi-directional iterator."""
    def __init__(self, quotes):
        self.quotes = quotes
        self.indices = list(range(len(self.quotes)))
        self.index = 0
        random.shuffle(self.indices)

    def __len__(self):
        return len(self.quotes)

    def _current(self):
        index = self.indices[self.index]

        quote = self.quotes[index]
        author = quote[0]
        title = quote[1]
        text = quote[2]

        if len(quote) > 3:
            text_id = quote[3]
        else:
            text_id = index

        return Quote(author, title, text, text_id)

    @property
    def database(self):
        return self.quotes.database

    @property
    def text_id(self):
        return self.indices[self.index]

    def next(self):
        self.index = (self.index + 1) % len(self.quotes)
        if self.index == 0:
            random.shuffle(self.indices)
        return self._current()

    def previous(self):
        self.index = (self.index - 1) % len(self.quotes)
        return self._current()


class Quotes(object):
    def __init__(self, quotes=None, database=None):
        self.quotes = quotes
        self.database = database

    def __len__(self):
        return len(self.quotes)

    def __getitem__(self, index):
        return self.quotes[index]

    def __setitem__(self, index, item):
        if not isinstance(item, list):
            raise ValueError("Expected a list")
        if len(item) != 3:
            raise ValueError("Expected a list of three strings")
        self.quotes[index] = item

    def random_iterator(self):
        return RandomIterator(self)

    @staticmethod
    def _database_filename():
        return pkg_resources.resource_filename("wpm", "data/examples.json.gz")

    @staticmethod
    def load_json(filename=None):
        if filename is None:
            filename = Quotes._database_filename()

        with codecs.open(filename, encoding="utf-8") as f:
            quotes = json.load(f)

        # Flatten
        out = []
        for text_id, quote in enumerate(quotes):
            author = quote["author"]
            title = quote["title"]
            text = quote["text"]
            text_id = int(quote.get("text_id", text_id))
            out.append((author, title, text, text_id))

        return Quotes(list(set(out)))

    @staticmethod
    def load(filename=None):
        if filename is None:
            filename = Quotes._database_filename()
            database = "default"
        else:
            database = os.path.splitext(os.path.basename(filename))[0]

        with gzip.open(filename) as f:
            quotes = json.load(f)
            quotes = tuple(map(tuple, quotes))
            return Quotes(quotes, database)

    def save(self, filename=None):
        if filename is None:
            filename = Quotes._database_filename()

        with gzip.open(filename, mode="wb") as f:
            json.dump(self.quotes, f)



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017 Christian Stigen Larsen

Distributed under the GNU GPL v3 or later. See the file LICENSE.txt for the
full license text. This software makes use of open source software.
"""

import argparse
import codecs
import math
import os
import sys
import wpm
import wpm.error
import wpm.game
import wpm.quotes
import wpm.stats

def parse_args():
    p = argparse.ArgumentParser(prog="wpm", epilog=wpm.__copyright__)

    p.add_argument("--load-json", metavar="FILENAME", default=None,
            help="""JSON file containing texts to train on.

The format is

[{"author": "...", "title": "...", "text": "..."}, ...]
""")
    p.add_argument("--load", metavar="FILENAME", default=None,
            help="A pure text file to train on.")

    p.add_argument("-V", "--version", default=False, action="store_true",
            help="Show program version")

    p.add_argument("--tabs", default=None, type=int,
            help="If set, expand tabs to this number of spaces")

    p.add_argument("--keyboard", default=None, type=str,
            help="Records WPM statistics under the given keyboard name")

    p.add_argument("-s", "--stats", default=False, action="store_true",
            help="Shows keyboard statistics")

    p.add_argument("--stats-file", default="~/.wpm.csv", type=str,
            help="File to record score history to (CSV format)")

    p.add_argument("--id", "-i", default=None, type=int,
            help="If specified, jumps to given text ID on start.")

    opts = p.parse_args()

    if opts.version:
        print("WPM v%s" % wpm.__version__)
        print(wpm.__copyright__)
        print("Distributed under the %s" % wpm.__license__)
        sys.exit(0)

    opts.stats_file = os.path.expanduser(opts.stats_file)
    return opts

def averages(games):
    wpms = []
    accs = []

    # TODO: Fix this
    for game in games:
        wpm = game[1]
        wpms.append(wpm)

        accuracy = game[2]
        accs.append(accuracy)

    average_wpm = sum(wpms) / len(wpms)
    average_acc = sum(accs) / len(accs)

    return average_wpm, average_acc

def stddevs(games):
    awpm, aacc = averages(games)
    n = len(games)

    if n <= 1:
        return 0, 0

    swpm = 0
    sacc = 0
    for game in games:
        wpm = game[1]
        acc = game[2]
        swpm += (wpm - awpm)**2.0
        sacc += (acc - aacc)**2.0

    return math.sqrt(swpm/(n-1)), math.sqrt(sacc/(n-1))

def main():
    opts = parse_args()
    quotes = []

    if not os.path.isfile(opts.stats_file):
        stats = wpm.stats.Stats(opts.keyboard)
    else:
        try:
            stats = wpm.stats.Stats.load(opts.stats_file)
        except Exception as e:
            print("Unsupported format. Renaming %s to %s" % (opts.stats_file,
                opts.stats_file + ".old"))
            os.rename(opts.stats_file, opts.stats_file + ".old")
            stats = wpm.stats.Stats()

        if opts.keyboard is not None:
            stats.keyboard = opts.keyboard

    if opts.load_json is not None:
        quotes = wpm.quotes.Quotes.load_json(opts.load_json)

    if opts.load is not None:
        with codecs.open(opts.load, encoding="utf-8") as f:
            text = f.read().replace("\r", "").rstrip()
        quotes.append(["", "", text])

        quotes = wpm.quotes.Quotes(quotes)

    if len(quotes) == 0 and not opts.stats:
        quotes = wpm.quotes.Quotes.load()

    if opts.stats:
        table = []
        for keyboard in sorted(stats.games.keys()):
            name = keyboard if keyboard is not None else "Unspecified"

            games = stats.games[keyboard]
            awpm, aacc = averages(games)
            swpm, sacc = stddevs(games)

            for last_n in [0, 10, 100, 1000]:
                if len(games) >= last_n:
                    if last_n == 0:
                        label = len(games)
                    else:
                        label = "n-%d" % last_n

                    awpm, aacc = averages(games[-last_n:])
                    swpm, sacc = stddevs(games[-last_n:])

                    table.append([name, label, awpm, swpm, 100.0*aacc, 100.0*sacc])

        width = max(max(len(e[0]) for e in table), 11)
        print("Keyboard      Games    WPM avg stddev     Accuracy avg stddev")
        print("-------------------------------------------------------------")
        for entry in table:
            label, count, wpm_avg, wpm_sd, acc_avg, acc_sd = entry
            print("%-*s   %5s      %5.1f  %5.1f            %4.1f%% %5.1f%%" %
                    (width, label, count, wpm_avg, wpm_sd, acc_avg, acc_sd))
        return

    try:
        with wpm.game.Game(quotes, stats) as game:
            game.set_tab_spaces(opts.tabs)
            if opts.id is not None:
                try:
                    game.jump_to(opts.id)
                except KeyError as e:
                    raise wpm.error.WpmError("Text ID %d not found" % opts.id)
                    sys.exit(1)
            game.run()
    except KeyboardInterrupt:
        game.stats.save(opts.stats_file)
        sys.exit(0)
    except wpm.error.WpmError as e:
        print(e)
        sys.exit(1)



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017 Christian Stigen Larsen

Distributed under the GNU GPL v3 or later. See the file LICENSE.txt for the
full license text. This software makes use of open source software.
"""

import wpm.commandline

if __name__ == "__main__":
    wpm.commandline.main()



# -*- encoding: utf-8 -*-

"""
This file is part of the wpm software.
Copyright 2017 Christian Stigen Larsen

Distributed under the GNU GPL v3 or later. See the file LICENSE.txt for the
full license text. This software makes use of open source software.
"""

import collections
import csv
import datetime
import os

class Timestamp(object):
    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

    @staticmethod
    def from_string(string):
        return datetime.datetime.strptime(string, Timestamp.DATETIME_FORMAT)

    @staticmethod
    def now():
        return datetime.datetime.utcnow()


class Stats(object):
    """Typing statistics"""

    def __init__(self, current_keyboard=None):
        self.keyboard = current_keyboard
        self.games = collections.defaultdict(list)

    def add(self, wpm, accuracy, text_id, database):
        race = 0
        rank = 1
        racers = 1

        self.games[self.keyboard].append((
            race,
            wpm,
            accuracy,
            rank,
            racers,
            text_id,
            Timestamp.now(),
            database))

    def average(self, keyboard=None, last_n=None):
        total = []

        for kbd in self.games.keys():
            if keyboard is not None:
                if kbd != keyboard:
                    continue
            for game in self.games[kbd]:
                total.append(game[1])

        if (last_n is not None) and len(total) >= last_n:
            total = total[-last_n:]

        if len(total) > 0:
            return sum(total) / len(total)
        else:
            return 0

    def __len__(self):
        return len(self.games)

    def __getitem__(self, key):
        return self.games[key]

    def keys(self):
        return self.games.keys()

    def values(self):
        return self.games.values()

    def items(self):
        return self.games.items()

    @staticmethod
    def load(filename):
        games = collections.defaultdict(list)
        current_keyboard = None

        with open(filename, "rt") as f:
            reader = csv.reader(f)
            for row in reader:
                race = int(row[0])
                wpm = float(row[1])
                accuracy = float(row[2])
                rank = int(row[3])
                racers = int(row[4])
                text_id = int(row[5])
                timestamp = Timestamp.from_string(row[6])
                database = row[7]
                keyboard = row[8]

                if keyboard not in games:
                    games[keyboard] = []
                games[keyboard].append((race, wpm, accuracy, rank, racers,
                    text_id, timestamp, database))
                current_keyboard = keyboard

        stats = Stats(current_keyboard)
        stats.games = games
        return stats

    def save(self, filename):
        """Writes game results to a CSV file compatible with the one from
        TypeRacer."""
        allgames = []
        for keyboard, games in self.items():
            if keyboard is None:
                keyboard = "Unspecified"

            for game in games:
                allgames.append(list(game) + [keyboard])

        by_time = lambda row: row[6]
        games = sorted(allgames, key=by_time)

        # Write to a temp file just in case we get an exception
        with open(filename + ".tmp", "wt") as f:
            writer = csv.writer(f)

            for race, game in enumerate(games):
                # The timestamp is the race number, so use that instead.
                game[0] = 1 + race
                writer.writerow(game)

        os.rename(filename + ".tmp", filename)



#! /usr/bin/env python
# -*- encoding: utf-8 -*-

"""
Measures your typing speed in words per minute (WPM).

This file is part of the wpm software.
Copyright 2017 Christian Stigen Larsen

Distributed under the GNU GPL v3 or later. See the file LICENSE.txt for the
full license text. This software makes use of open source software.
"""

import curses
import curses.ascii
import locale
import os
import sys
import time
import wpm.error

def word_wrap(s, w):
    """Returns lengths of lines that can be printed without wrapping."""
    lengths = []
    while len(s) > w:
        end = s[:w+1].rindex(" ")

        # We can't divide the input nicely, so just display it as-is
        if end == -1:
            return [len(s)]

        lengths.append(end)
        s = s[end+1:]

    if len(s) > 0:
        lengths.append(len(s))

    return lengths

def screen_coords(lens, pos):
    for y, l in enumerate(lens):
        if pos <= l:
            break
        pos -= (l+1)
    return pos, y

class Screen(object):
    def __init__(self):
        # Make delay slower
        os.environ.setdefault("ESCDELAY", "15")
        locale.setlocale(locale.LC_ALL, "")
        self.screen = curses.initscr()
        self.screen.nodelay(True)

        if curses.LINES < 12:
            curses.endwin()
            raise wpm.error.WpmError(
                    "wpm requires at least 12 lines in your display")

        if curses.COLS < 51:
            curses.endwin()
            raise wpm.error.WpmError(
                    "wpm requires at least 51 columns in your display")

        self.screen.keypad(True)
        curses.noecho()
        curses.cbreak()
        curses.start_color()

        if os.getenv("TERM").endswith("256color"):
            bg = 233

            # Incorrect
            curses.init_pair(1, 197, 52)

            # Status
            curses.init_pair(2, 51, 24)

            # Done text
            curses.init_pair(3, 240, bg)

            # Normal text
            curses.init_pair(4, 195, bg)

            # UNUSED
            curses.init_pair(5, 244, bg)

            # Author
            curses.init_pair(6, 240, bg)

            # Edit text and info
            curses.init_pair(7, 244, 233)

            # Background color
            curses.init_pair(8, bg, bg)

            # Score highlight
            curses.init_pair(9, 230, 197)

        else:
            bg = curses.COLOR_BLACK

            # Incorrect
            curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_RED)

            # Status
            curses.init_pair(2, curses.COLOR_CYAN, bg)

            # Done text
            curses.init_pair(3, curses.COLOR_WHITE, bg)

            # Normal text
            curses.init_pair(4, curses.COLOR_WHITE, bg)

            # UNUSED
            curses.init_pair(5, curses.COLOR_WHITE, bg)

            # Author
            curses.init_pair(6, curses.COLOR_WHITE, bg)

            # Edit text and info
            curses.init_pair(7, curses.COLOR_WHITE, bg)

            # Background color
            curses.init_pair(8, bg, bg)

            # Score highlight
            curses.init_pair(9, curses.COLOR_YELLOW, curses.COLOR_RED)


        self.window = curses.newwin(curses.LINES, curses.COLS, 0, 0)
        self.window.keypad(True)
        self.window.timeout(20)
        self.window.bkgd(" ", curses.color_pair(8))

    def is_escape(self, key):
        if len(key) == 1:
            return ord(key) == curses.ascii.ESC
        return False

    def is_backspace(self, key):
        if len(key) > 1:
            return key == "KEY_BACKSPACE"
        elif ord(key) in (curses.ascii.BS, curses.ascii.DEL):
            return True
        return False

    def get_key(self):
        # Install a suitable get_key based on Python version
        if sys.version_info[0:2] >= (3, 3):
            self.get_key = self.getkey_py33
        else:
            self.get_key = self.getkey_py27
        return self.get_key()

    def getkey_py33(self):
        try:
            # Curses in Python 3.3 handles unicode via get_wch
            key = self.window.get_wch()
            if type(key) == int:
                if key == curses.KEY_LEFT:
                    return "KEY_LEFT"
                elif key == curses.KEY_RIGHT:
                    return "KEY_RIGHT"
                elif key == curses.KEY_RESIZE:
                    return "KEY_RESIZE"
                else:
                    return None
            return key
        except curses.error:
            return None
        except KeyboardInterrupt:
            raise

    def getkey_py27(self):
        try:
            return self.window.getkey()
        except KeyboardInterrupt:
            raise
        except curses.error:
            return None

    def column(self, y, x, width, text, attr=None, left=True):
        lengths = word_wrap(text, width)
        for y, length in enumerate(lengths, y):
            if left:
                self.window.addstr(y, x, text[:length].encode("utf-8"), attr)
            else:
                self.window.addstr(y, x - length, text[:length].encode("utf-8"), attr)
            text = text[1+length:]
        return len(lengths)

    def update(self, browse, head, quote, position, incorrect, author, title,
            typed, wpm, average):
        cols = curses.COLS
        lengths = word_wrap(quote, cols - 1)
        sx, sy = screen_coords(lengths, position)
        h = len(lengths)

        # Show header
        self.window.addstr(0, 0, head + " "*(cols - len(head)),
                curses.color_pair(2))

        if browse:
            # Display quote
            color = curses.color_pair(4 if browse == 1 else 3)
            for y, length in enumerate(lengths, 2):
                self.window.addstr(y, 0, quote[:length], color)
                quote = quote[1+length:]

            # Show author
            credit = u"— %s, %s" % (author, title)
            self.cheight = 4 + h + self.column(3+h, cols - 10, cols//2, credit,
                    curses.color_pair(6), False)
            if browse >= 2:
                typed = "You scored %.1f wpm%s " % (wpm, "!" if wpm > average
                        else ".")
            else:
                typed = ""
            typed += "Use arrows/space to browse, esc to quit, or start typing."
        elif position < len(quote):
            color = curses.color_pair(3 if incorrect == 0 else 1)
            typed = "> " + typed

            if position + incorrect < len(quote):
                sx, sy = screen_coords(lengths, position + incorrect - 1)
                self.window.chgat(2 + sy, max(sx, 0), 1, color)

                sx, sy = screen_coords(lengths, position + incorrect + 1)
                self.window.chgat(2 + sy, sx, curses.color_pair(4))

        # Show typed text
        if self.cheight < curses.LINES:
            self.window.move(self.cheight, 0)
            self.window.clrtoeol()
            self.window.addstr(self.cheight, 0, typed, curses.color_pair(7))
        if browse > 1:
            # If done, highlight score
            self.window.chgat(self.cheight, 11,
                len(str("%.1f" % wpm)), curses.color_pair(9))

        # Move cursor to current position in text before refreshing
        if browse < 1:
            sx, sy = screen_coords(lengths, position + incorrect)
            self.window.move(2 + sy, min(sx, cols - 1))
        else:
            self.window.move(2, 0)

        self.window.refresh()

    def clear(self):
        self.window.clear()

    def deinit(self):
        curses.nocbreak()
        self.screen.keypad(False)
        curses.echo()
        curses.endwin()

class Game(object):
    def __init__(self, quotes, stats):
        self.stats = stats
        self.average = self.stats.average(self.stats.keyboard, last_n=10)
        self.tab_spaces = None

        # Stats
        self.position = 0
        self.incorrect = 0
        self.total_incorrect = 0

        self.start = None
        self.stop = None

        self._edit = ""
        self.num_quotes = len(quotes)
        self.quotes = quotes.random_iterator()
        self.quote = self.quotes.next()

        self.screen = Screen()

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.screen.deinit()
        if type is not None:
            return False
        else:
            return self

    def jump_to(self, text_id):
        """Skips until then given text id is found."""
        pos = 0
        while self.quote.text_id != text_id:
            self.quote = self.quotes.next()
            pos += 1
            if pos > self.num_quotes:
                raise KeyError("Text ID not found: %s" % text_id)

    def set_tab_spaces(self, spaces):
        self.tab_spaces = spaces

    def mark_finished(self):
        self.stop = time.time()
        self.stats.add(self.wpm(self.elapsed), self.accuracy,
                self.quote.text_id, self.quotes.database)
        self.average = self.stats.average(self.stats.keyboard, last_n=10)

    def run(self):
        while True:
            is_typing = self.start is not None and self.stop is None

            browse = int(not is_typing)
            if self.stop is not None:
                browse = 2

            self.screen.update(browse, self.get_stats(self.elapsed),
                    self.quote.text, self.position, self.incorrect,
                    self.quote.author, self.quote.title, self._edit,
                    self.wpm(self.elapsed), self.average)

            key = self.screen.get_key()
            if key is not None:
                self.handle_key(key)

    def wpm(self, elapsed):
        """Words per minute."""
        if self.start is None:
            return 0
        else:
            return min((60.0 * self.position / 5.0) / elapsed, 999)

    def cps(self, elapsed):
        """Characters per second."""
        if self.start is None:
            return 0
        else:
            return min(float(self.position) / elapsed, 99)

    @property
    def elapsed(self):
        """Elapsed game round time."""
        if self.start is None:
            # Typing has not started
            return 0
        elif self.stop is None:
            # Currently typing
            return time.time() - self.start
        else:
            # Done typing
            return self.stop - self.start

    @property
    def accuracy(self):
        if self.start is None:
            return 0
        else:
            n = len(self.quote.text)
            return float(n) / (n + self.total_incorrect)

    def get_stats(self, elapsed):
        keyboard = self.stats.keyboard
        if keyboard is None:
            keyboard = "Unspecified"

        return "%5.1f wpm %4.1f cps %5.2fs %5.1f%% acc %5.1f avg wpm - %s" % (
                self.wpm(elapsed),
                self.cps(elapsed),
                elapsed,
                100.0*self.accuracy,
                self.average,
                keyboard)

    def reset(self, direction=0):
        self.start = None
        self.stop = None

        self.position = 0
        self.incorrect = 0
        self.total_incorrect = 0

        self._edit = ""

        if direction:
            if direction > 0:
                self.quote = self.quotes.next()
            else:
                self.quote = self.quotes.previous()
            self.screen.clear()

    def resize(self):
        y, x = self.screen.window.getmaxyx()
        self.screen.clear()
        curses.resizeterm(y, x)

    def handle_key(self, key):
        if key == "KEY_RESIZE":
            self.resize()
            return

        # Browse mode
        if self.start is None or (self.start is not None and self.stop is not
                None):
            if key in (" ", "KEY_LEFT", "KEY_RIGHT"):
                self.reset(direction=-1 if key == "KEY_LEFT" else 1)
                return
            elif self.screen.is_escape(key):
                # Exit program
                raise KeyboardInterrupt()

        if self.screen.is_escape(key):
            self.reset()
            return

        if self.screen.is_backspace(key):
            if self.incorrect > 0:
                self.incorrect -= 1
                self._edit = self._edit[:-1]
            elif len(self._edit) > 0:
                self.position -= 1
                self._edit = self._edit[:-1]
            return

        # Try again?
        if self.stop is not None:
            self.reset()
            self.screen.clear()
            self.screen.update(1, self.get_stats(self.elapsed),
                    self.quote.text, self.position, self.incorrect,
                    self.quote.author, self.quote.title, self._edit,
                    self.wpm(self.elapsed), self.average)

        # Start recording upon first ordinary key press
        if self.start is None:
            self.start = time.time()

        if key == curses.KEY_ENTER:
            key = "\n"
        elif key == "\t" and self.tab_spaces is not None:
            key = " "*self.tab_spaces

        # Did the user strike the correct key?
        if self.incorrect == 0 and self.quote.text[self.position] == key:
            self.position += 1

            # Reset edit buffer on a correctly finished word
            if key == " " or key == "\n":
                self._edit = ""
            else:
                self._edit += key

            # Finished typing?
            if self.position == len(self.quote.text):
                self.mark_finished()
        elif self.incorrect + self.position < len(self.quote.text):
            self.incorrect += 1
            self.total_incorrect += 1
            if key == "\n":
                key = " "
            self._edit += key



from setuptools import setup
import os
import runpy

def get_version():
    filename = os.path.join(os.path.dirname(__file__), "wpm",
            "__init__.py")
    var = runpy.run_path(filename)
    return var["__version__"]

_VERSION = get_version()

setup(
    name="wpm",
    scripts=["scripts/wpm"],
    version=_VERSION,
    description="Console app for measuring typing speed in words per minute (WPM)",
    author="Christian Stigen Larsen",
    author_email="csl@csl.name",
    packages=["wpm"],
    package_dir={"wpm": "wpm"},
    package_data={"wpm": ["data/examples.json.gz"]},
    include_package_data=True,
    url="https://github.com/cslarsen/wpm",
    download_url="https://github.com/cslarsen/wpm/tarball/v%s" % _VERSION,
    license="https://www.gnu.org/licenses/gpl-3.0.html",
    long_description=open("README.rst").read(),
    zip_safe=True,
    keywords=["wpm", "typing", "typist"],
    platforms=["unix", "linux", "osx", "cygwin", "win32"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.1",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
)
